def _get_path(component):
    """Get the vocabulary path from the component."""
    for coding in component.get("code", {}).get("coding", []):
        if coding.get("system", "") == "http://fhir-aggregator.org/fhir/CodeSystem/vocabulary/path":
            return coding.get("code", None)


def _get_coding(component):
    """Get the vocabulary codeable from the component."""
    for coding in component.get("code", {}).get("coding", []):
        if coding.get("system", "") != "http://fhir-aggregator.org/fhir/CodeSystem/vocabulary/path":
            return coding


def vocabulary_simplifier(bundle) -> list[dict]:
    """Simplify the vocabulary bundle."""
    df = []
    resources = {f'{r["resource"]["resourceType"]}/{r["resource"]["id"]}': r["resource"] for r in bundle.get("entry", [])}
    for _id, resource in resources.items():
        if resource["resourceType"] != "Observation":
            continue
        focus = next(iter(resource.get("focus", [])), None)
        assert focus, f"No focus found for Observation {resource['id']}"
        focus_reference = focus.get("reference", None)
        research_study = resources.get(focus_reference, None)
        assert research_study, f"No research_study reference found for Observation {resource['id']} {focus_reference}"
        for component in resource.get("component", []):
            path = _get_path(component)
            coding = _get_coding(component)
            item = {
                "research_study_identifiers": ",".join([i.get("value", "") for i in research_study.get("identifier", [])]),
                "path": path,
            }
            if path.endswith(".extension"):
                item.update(
                    {
                        "code": coding.get("code", None) if coding.get("code", None) != "range" else None,
                        "display": coding.get("display", None) if coding.get("display", None) != "range" else None,
                        "system": None,
                        "extension_url": coding.get("system", None),
                    }
                )
            else:
                item.update(
                    {
                        "code": coding.get("code", None),
                        "display": coding.get("display", None),
                        "system": coding.get("system", None),
                        "extension_url": None,
                    }
                )

            if "valueInteger" in component:
                item["count"] = component["valueInteger"]
            else:
                item["count"] = None
            if "valueRange" in component:
                item["low"] = component["valueRange"].get("low", {}).get("value", None)
                item["high"] = component["valueRange"].get("high", {}).get("value", None)
            else:
                item["low"] = None
                item["high"] = None
            item.update(
                {
                    "research_study_title": research_study.get("title", None),
                    "research_study_description": research_study.get("description", None),
                    "observation": f'Observation/{resource["id"]}',
                    "research_study": f'ResearchStudy/{research_study["id"]}',
                }
            )
            df.append(item)
    return df



# Initialize the fhir_query package
import asyncio
import json
import logging
import sqlite3
import tempfile
from collections import defaultdict
from typing import Any, Optional, Callable

import httpx
from dotty_dict import dotty
from halo import Halo

UNKNOWN_CATEGORY = {"coding": [{"system": "http://snomed.info/sct", "code": "261665006", "display": "Unknown"}]}


def setup_logging(debug: bool, log_file: str) -> None:
    """
    Set up logging configuration.

    Args:
        debug (bool): Enable debug mode if True.
        log_file (str): Path to the log file.
    """
    log_level = logging.DEBUG if debug else logging.INFO
    file_handler = logging.FileHandler(log_file)
    logging.basicConfig(level=log_level, handlers=[file_handler])

    # Configure httpx logger
    httpx_logger = logging.getLogger("httpx")
    httpx_logger.setLevel(log_level)
    httpx_logger.addHandler(file_handler)


class ResourceDB:
    def __init__(self, db_path: str = ":memory:"):
        """
        Initialize the ResourceDB instance and create the resources table if it doesn't exist.

        Args:
            db_path (str): Path to the SQLite database file (default is in-memory database).
        """
        self.db_path = db_path
        self.connection = sqlite3.connect(db_path)
        self._logged_already: list[str] = []
        self.adds_counters: dict[str, int] = defaultdict(int)
        self._initialize_table()

    def _initialize_table(self) -> None:
        """
        Create the 'resources' table if it doesn't already exist.
        """
        with self.connection:
            self.connection.execute(
                """
                CREATE TABLE IF NOT EXISTS resources (
                    id VARCHAR NOT NULL,
                    resource_type VARCHAR NOT NULL,
                    key VARCHAR NOT NULL,
                    resource JSON NOT NULL,
                    PRIMARY KEY (id, resource_type)
                )
                """
            )
            self.connection.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_resources_key
                ON resources (key)
                """
            )

    def add(self, resource: dict[str, Any]) -> None:
        """
        Add a resource to the 'resources' table.

        Args:
            resource (dict): A dictionary with 'id', 'resourceType', and other fields.
        """
        if "id" not in resource or "resourceType" not in resource:
            raise ValueError("Resource must contain 'id' and 'resourceType' fields.")

        try:
            with self.connection:
                self.connection.execute(
                    """
                    INSERT INTO resources (id, resource_type, key, resource)
                    VALUES (?, ?, ?, ?)
                """,
                    (
                        resource["id"],
                        resource["resourceType"],
                        f'{resource["resourceType"]}/{resource["id"]}',
                        json.dumps(resource),
                    ),
                )
                self.adds_counters[resource["resourceType"]] += 1
        except sqlite3.IntegrityError as e:
            if "UNIQUE constraint failed" in str(e):
                pass
            else:
                raise

    def all_keys(self, resource_type: str) -> list[Any]:
        """
        Retrieve all (id, resource_type) tuples for a given resource_type.

        Args:
            resource_type (str): The resource type to filter by.

        Returns:
            list: A list of tuples (id, resource_type).
        """
        with self.connection:
            cursor = self.connection.execute(
                """
                SELECT id, resource_type
                FROM resources
                WHERE resource_type = ?
            """,
                (resource_type,),
            )
            return cursor.fetchall()

    def all_resources(self, resource_type: str) -> list[dict[str, Any]]:
        """
        Retrieve all resource dicts for a given resource_type.

        Args:
            resource_type (str): The resource type to filter by.

        Returns:
            list: A list of dicts.
        """
        with self.connection:
            cursor = self.connection.execute(
                """
                SELECT resource
                FROM resources
                WHERE resource_type = ?
            """,
                (resource_type,),
            )
            return [json.loads(row[0]) for row in cursor.fetchall()]

    def count_resource_types(self) -> dict[str, Any]:
        """
        Count the number of resources for each resource_type.

        Returns:
            dict: A dictionary with resource_type as keys and counts as values.
        """
        with self.connection:
            cursor = self.connection.execute(
                """
                   SELECT resource_type, COUNT(*)
                   FROM resources
                   GROUP BY resource_type
               """
            )
            return {row[0]: row[1] for row in cursor.fetchall()}

    def close(self) -> None:
        """
        Close the database connection.
        """
        self.connection.close()

    def aggregate(self, ignored_edges=None) -> dict:
        """
        Aggregate metadata counts resourceType(count)-count->resourceType(count).

        Args:
            ignored_edges (list): List of edges to ignore during aggregation.

        Returns:
            dict: Aggregated metadata counts.
        """
        if ignored_edges is None:
            ignored_edges = []
        nested_dict: Callable[[], defaultdict[str, defaultdict]] = lambda: defaultdict(defaultdict)

        count_resource_types = self.count_resource_types()

        summary = nested_dict()

        for resource_type in count_resource_types:

            resources = self.all_resources(resource_type)

            for _ in resources:

                if "count" not in summary[resource_type]:
                    summary[resource_type]["count"] = 0
                summary[resource_type]["count"] += 1

                # refs = nested_lookup("reference", _)
                refs = find_key_with_path(_, "reference", ignored_keys=ignored_edges)
                for match in refs:
                    path, ref = match.values()

                    # A codeable reference is an object with a codeable concept and a reference
                    if isinstance(ref, dict):
                        ref = ref["reference"]
                    ref_resource_type = ref.split("/")[0]
                    if "references" not in summary[resource_type]:
                        summary[resource_type]["references"] = nested_dict()

                    # # only count references to resources that are not ResearchStudy
                    # if ref_resource_type == 'ResearchStudy' and resource_type != 'ResearchSubject':
                    #     continue
                    # if set([resource_type, ref_resource_type]) == set(['Group', 'Patient']):
                    #     pass
                    dst = summary[resource_type]["references"][ref_resource_type]
                    if "count" not in dst:
                        dst["count"] = 0
                    dst["count"] += 1

        return summary


class GraphDefinitionRunner(ResourceDB):
    """
    A class to parse a FHIR GraphDefinition and execute the queries defined in its links.

    See
    https://www.devdays.com/wp-content/uploads/2021/12/Rene-Spronk-GraphDefinition-_-DevDays-2019-Amsterdam-1.pdf
    """

    def __init__(self, fhir_base_url: str, db_path: Optional[str] = None, debug: Optional[bool] = False):
        """
        Initializes the GraphDefinitionRunner.

        Args:
            fhir_base_url (str): Base URL of the FHIR server.
            db_path (Optional[str]): Path to the SQLite database file. Defaults to a temporary file.
            debug (Optional[bool]): Enable debug mode if True. Defaults to False.
        """
        if not db_path:
            # initializes the ResourceDB to a temporary file
            db_path = tempfile.NamedTemporaryFile(delete=False).name

        super().__init__(db_path)
        self.fhir_base_url = fhir_base_url
        self.max_requests = 10
        self.debug = debug
        self.recurse_count = 0

    async def fetch_graph_definition(self, graph_definition_id: str) -> Any:
        """
        Fetches the GraphDefinition resource from the FHIR server.

        Args:
            graph_definition_id (str): ID of the GraphDefinition resource.

        Returns:
            dict: Parsed JSON response of the GraphDefinition resource.
        """
        async with httpx.AsyncClient() as client:
            url = f"{self.fhir_base_url}/GraphDefinition/{graph_definition_id}"
            response = await client.get(url)
            response.raise_for_status()
            return response.json()

    async def execute_query(self, query_url: str, spinner: Halo = None, page_count: int = 0) -> list[dict[str, Any]]:
        """
        Executes a FHIR query for a given URL, returns all pages as a list of resources.

        Args:
            query_url (str): Fully constructed query URL.
            spinner (Halo): Spinner object to show progress. Defaults to None.
            page_count (int): The current page count. Defaults to 0.

        Returns:
            list: A list of resources from the query result.
        """
        retry = 0
        max_retry = 3
        log_every_n_pages = 10
        while retry < max_retry:
            async with httpx.AsyncClient() as client:
                try:
                    if self.debug:
                        print(f"Querying: {query_url}")
                    response = await client.get(query_url, timeout=300)
                    response.raise_for_status()
                    page_count += 1
                    query_result = response.json()
                    resources = []
                    next_link = [link for link in query_result.get("link", []) if link["relation"] == "next"]
                    entries = query_result.get("entry", [])
                    for entry in entries:
                        self.add(entry["resource"])
                        resources.append(entry["resource"])
                    if next_link:
                        if spinner and page_count % log_every_n_pages == 0:
                            estimated_number_of_pages = "unknown"
                            resource_type = "unknown"
                            if entries:
                                resource_type = entries[0]["resource"]["resourceType"]
                            if "total" in query_result:
                                estimated_number_of_pages = query_result["total"] // len(entries)
                            spinner.info(f"Fetching {resource_type} page {page_count} of {estimated_number_of_pages}")
                        for entry in await self.execute_query(next_link[0]["url"], spinner=spinner, page_count=page_count):
                            resources.append(entry)

                    if self.debug:
                        print(f"Query result: {len(resources)}")
                    return resources

                except httpx.ReadTimeout as e:
                    if retry == max_retry:
                        logging.warning(f"ReadTimeout: {e} sleeping for 5 seconds. Retry: {retry}")
                    await asyncio.sleep(5)
                    retry += 1
                except httpx.ConnectTimeout as e:
                    if retry == max_retry:
                        logging.warning(f"ConnectTimeout: {e} sleeping for 5 seconds. Retry: {retry}")
                    await asyncio.sleep(5)
                    retry += 1
                except httpx.RemoteProtocolError as e:
                    if retry == max_retry:
                        logging.warning(f"RemoteProtocolError: {e} sleeping for 5 seconds. Retry: {retry}")
                    await asyncio.sleep(5)
                    retry += 1

        return []

    async def process_link(self, link, parent_resources, visited, spinner):
        """
        Processes a single link in the GraphDefinition.

        Args:
            link (dict): The link to process.
            parent_resources (list): List of parent resources.
            visited (set): Set of visited node-resource combinations to prevent cycles.
            spinner (Halo): Spinner object to show progress.
        """
        target_id = link["targetId"]
        source_id = link["sourceId"]
        if "params" in link:
            params = link["params"]
            current_path = set()
            for _ in parent_resources:
                if _["resourceType"] == source_id:
                    key = (_["resourceType"], _["id"], target_id)
                    if key not in visited:
                        visited.add(key)
                        parent = dotty({_["resourceType"]: _})
                        assert "path" in link, f"Path is required for {link}"
                        path = link["path"]
                        if path not in parent:
                            continue
                        _path = parent[path]
                        if path.endswith(".id"):
                            _path = source_id + "/" + _path
                        if "_id={path}" in params and "/" in _path:
                            _path = _path.split("/")[-1]
                        current_path.add(_path)
            if not current_path:
                if spinner:
                    spinner.fail(f"Could not find any resources for {source_id}->{target_id} link: {link}")
                return
            if spinner:
                spinner.info(
                    text=f"Processing link: {link['targetId']}/{link['params']} with {len(current_path)} {link['sourceId']}(s)"
                )
            _current_path = list(current_path)
            chunk_size = 40
            chunks = [_current_path]
            if len(_current_path) > chunk_size:
                chunks = [_current_path[i : i + chunk_size] for i in range(0, len(_current_path), chunk_size)]
            tasks = []
            for chunk in chunks:
                _params = params.replace("{path}", ",".join(chunk))
                query_url = f"{self.fhir_base_url}/{target_id}?{_params}"
                tasks.append(asyncio.create_task(self.execute_query(query_url, spinner=spinner)))
                if len(tasks) >= self.max_requests:
                    await asyncio.gather(*tasks)
                    tasks = []
            await asyncio.gather(*tasks)
        else:
            logging.debug(f"No `params` property found in link. {link} continuing")

        if spinner:
            spinner.clear()
            spinner.succeed(f"Processed link: {link['targetId']}/{link['params']}")

    async def process_links(
        self,
        parent_target_id: str,
        parent_resources: list[dict[str, Any]],
        graph_definition: dict[str, Any],
        visited: set[tuple[Any, Any, Any]],
        spinner: Halo,
    ) -> None:
        """
        Processes all links in the GraphDefinition for the given resource.

        Args:
            parent_target_id (str): The resource_type of the parent resource.
            parent_resources (list): Resources returned from the last query, can have multiple resource types.
            graph_definition (dict): The entire GraphDefinition resource.
            visited (set): Set of visited node-resource combinations to prevent cycles.
            spinner (Halo): Spinner object to show progress.
        """
        links = [link for link in graph_definition.get("link", []) if link.get("sourceId") == parent_target_id]
        if spinner:
            spinner.info(f"Processing {len(links)} links for {parent_target_id} in parallel.")
        tasks = [self.process_link(link, parent_resources, visited, spinner) for link in links]
        await asyncio.gather(*tasks)

    async def run(
        self,
        graph_definition: dict[str, Any],
        path: str,
        spinner: Halo,
    ) -> None:
        """
        Runs the GraphDefinition queries starting from the specified resource.

        Args:
            graph_definition (dict): The GraphDefinition resource.
            path (str): Path to query the FHIR server and pass to the GraphDefinition.
            spinner (Halo): Spinner object to show progress.
        """
        visited: set[tuple[Any, Any, Any]] = set()

        if path:
            url = self.fhir_base_url + path
            assert self.recurse_count == 0, "Should not call this twice"
            if spinner:
                spinner.info(text=f"Fetching {url}")
            self.recurse_count += 1
            parent_resources = await self.execute_query(url, spinner=spinner)
            if spinner:
                spinner.clear()
        else:
            parent_resources = []

        if len(parent_resources) == 0:
            if spinner:
                spinner.fail("No resources found")

        parent_resource_types = self.count_resource_types().keys()
        processed_links = []
        while True:
            parallelize = defaultdict(list)
            for link in graph_definition["link"]:
                if link in processed_links:
                    continue
                if link["sourceId"] in parent_resource_types:
                    processed_links.append(link)
                    parallelize[link["sourceId"]].append(link)

            if not parallelize:
                break

            tasks = []
            for source_id, links in parallelize.items():
                if spinner:
                    spinner.info(text=f"Processing {source_id} with {len(parent_resources)} resources")

                # create a sub graph definition for the with only the links for the current source_id
                parent_resources = self.all_resources(source_id)
                _sub_graph_definition = {"link": links}

                tasks.append(
                    self.process_links(
                        parent_target_id=source_id,
                        parent_resources=parent_resources,
                        graph_definition=_sub_graph_definition,
                        visited=visited,
                        spinner=spinner,
                    )
                )

            await asyncio.gather(*tasks)
            parent_resource_types = self.count_resource_types().keys()


def tree() -> defaultdict:
    """A recursive defaultdict."""
    return defaultdict(tree)


class VocabularyRunner:
    """
    A class to fetch and collect vocabularies from a FHIR server.

    Args:
        fhir_base_url (str): Base URL of the FHIR server.
    """

    def __init__(self, fhir_base_url: str):
        """
        Initialize the VocabularyRunner instance.

        Args:
            fhir_base_url (str): Base URL of the FHIR server.
        """
        self.fhir_base_url = fhir_base_url

    async def fetch_resource(self, resource_type: str, spinner: Halo = None) -> dict[str, dict[Any, Any]]:
        """
        Fetch resources of a given type from the FHIR server.

        Args:
            resource_type (str): The type of resource to fetch.
            spinner (Halo, optional): A Halo spinner object to show progress. Defaults to None.

        Returns:
            dict: A dictionary with resource type as keys and counts as values.
        """
        counts: dict = {resource_type: {}}
        category_counts = counts[resource_type]
        timeout = httpx.Timeout(10.0, connect=60.0)
        async with httpx.AsyncClient(timeout=timeout) as client:
            page_count = 1
            url = f"{self.fhir_base_url}/{resource_type}?_count=1000&_total=accurate&_elements=category,code,type"
            while url:
                if spinner:
                    spinner.text = f"Fetching {resource_type} page {page_count}"
                response = await client.get(url)
                response.raise_for_status()
                page_count += 1
                data = response.json()
                for entry in data.get("entry", []):
                    resource = entry["resource"]
                    code = resource.get("code", resource.get("type", None))
                    if not code:
                        code = UNKNOWN_CATEGORY
                    for category in resource.get("category", [UNKNOWN_CATEGORY]):
                        for category_coding in category.get("coding", []):
                            assert "display" in category_coding, f"No 'display' property in coding: {category_coding}"
                            if category_coding["display"] not in category_counts:
                                category_counts[category_coding["display"]] = {}

                            code_counts = category_counts[category_coding["display"]]
                            for code_coding in code.get("coding", []):
                                assert "display" in code_coding, f"No 'display' property in coding: {code_coding}"
                                if code_coding["display"] not in code_counts:
                                    code_counts[code_coding["display"]] = 0
                                code_counts[code_coding["display"]] += 1
                next_link = next((link["url"] for link in data.get("link", []) if link["relation"] == "next"), None)
                if next_link:
                    assert "write-fhir" not in next_link, f"Found write-fhir in from {url} next link: {next_link}"
                url = next_link
        return counts

    async def collect(self, resource_types: list[str], spinner: Halo = None) -> list:
        """
        Collect vocabularies from the specified resource types.

        Args:
            resource_types (list[str]): A list of resource types to collect vocabularies from.
            spinner (Halo, optional): A Halo spinner object to show progress. Defaults to None.

        Returns:
            list: A list of dictionaries with resource type as keys and counts as values.
        """
        tasks = []

        for resource_type in resource_types:
            tasks.append(asyncio.create_task(self.fetch_resource(resource_type, spinner)))

        results = await asyncio.gather(*tasks)
        return [_ for _ in results]


def find_key_with_path(data, key_to_find, ignored_keys=None):
    """
    Traverse the dictionary and find all occurrences of a given key.
    Returns a list of dictionaries containing the path and value for each match.
    Paths containing keys in the ignored_keys list are skipped.

    :param data: The input dictionary or list to search.
    :param key_to_find: The key to look for in the data structure.
    :param ignored_keys: A list of keys to ignore during traversal.
    :return: A list of dictionaries with 'path' and 'value' for each match.
    """
    if ignored_keys is None:
        ignored_keys = []

    results = []

    def recursive_search(d, current_path=None):
        if current_path is None:
            current_path = []
        if isinstance(d, dict):
            for key, value in d.items():
                new_path = current_path + [key]

                if key in ignored_keys:
                    continue  # Skip paths containing ignored keys

                if key == key_to_find:
                    found_ignored_key_in_extension = False
                    if "extension" in new_path:
                        extension_url = get_value_from_path(data, new_path[:-2] + ["url"])
                        if extension_url:
                            for k in ignored_keys:
                                if k in extension_url:
                                    found_ignored_key_in_extension = True
                                    break
                    if found_ignored_key_in_extension:
                        continue
                    results.append({"path": new_path, "value": value})
                recursive_search(value, new_path)
        elif isinstance(d, list):
            for index, item in enumerate(d):
                new_path = current_path + [index]
                recursive_search(item, new_path)

    recursive_search(data)
    return results


def get_value_from_path(data, path):
    """
    Retrieve a value from a nested dictionary or list using a path array.

    :param data: The nested dictionary or list to retrieve the value from.
    :param path: A list representing the path to the desired value.
    :return: The value at the specified path, or None if the path is invalid.
    """
    try:
        for key in path:
            data = data[key]
        return data
    except (KeyError, IndexError, TypeError):
        return None



# tests/unit/test_fhir_server.py
import pathlib

import httpx
import pytest
from click.testing import CliRunner

from fhir_query.cli import main
from fhir_query.dataframer import Dataframer
from fhir_query.visualizer import visualize_aggregation


@pytest.mark.usefixtures("mock_fhir_server")
def test_get_patient() -> None:
    response = httpx.get("http://testserver/Patient/123")
    assert response.status_code == 200
    assert response.json() == {"resourceType": "Patient", "id": "123"}


@pytest.mark.usefixtures("mock_fhir_server")
def test_get_nonexistent_patient() -> None:
    response = httpx.get("http://testserver/Patient/999")
    assert response.status_code == 404
    assert response.json() == {"error": "Not found"}


@pytest.mark.asyncio
@pytest.mark.usefixtures("mock_fhir_server")
@pytest.mark.httpx_mock(can_send_already_matched_responses=True)
def test_runner(tmp_path: str) -> None:
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(
        main,
        [
            "--fhir-base-url",
            "http://testserver",
            "--path",
            "/ResearchStudy?_id=123",
            "--db-path",
            f"{tmp_path}/fhir-query.sqlite",
            "--graph-definition-file-path",
            "tests/fixtures/ResearchStudyGraph.yaml",
            "--log-file",
            f"{tmp_path}/fhir-query.log",
            "--debug",
        ],
    )
    print(result.stderr)
    print(result.stdout)
    assert result.exit_code == 0, "CLI command failed"
    assert "research-study-graph" in result.stderr, result.stderr
    assert "database available at:" in result.stderr

    assert pathlib.Path(f"{tmp_path}/fhir-query.sqlite").exists()
    assert pathlib.Path(f"{tmp_path}/fhir-query.log").exists()

    # test the database

    db = Dataframer(f"{tmp_path}/fhir-query.sqlite")
    count_resource_types = db.count_resource_types()
    print(count_resource_types)
    assert count_resource_types == {
        "Condition": 1,
        "DocumentReference": 2,
        "Group": 1,
        "ImagingStudy": 1,
        "Medication": 1,
        "MedicationAdministration": 1,
        "Observation": 1,
        "Patient": 3,
        "Procedure": 1,
        "ResearchStudy": 1,
        "ResearchSubject": 1,
        "ServiceRequest": 1,
        "Specimen": 3,
    }

    aggregated = db.aggregate()
    aggregated_keys = sorted(aggregated.keys())
    print(aggregated_keys)
    assert aggregated_keys == [
        "Condition",
        "DocumentReference",
        "Group",
        "ImagingStudy",
        "Medication",
        "MedicationAdministration",
        "Observation",
        "Patient",
        "Procedure",
        "ResearchStudy",
        "ResearchSubject",
        "ServiceRequest",
        "Specimen",
    ]

    assert aggregated["Patient"]["count"] == 3
    assert aggregated["Specimen"]["count"] == 3
    assert aggregated["Specimen"]["references"]["Patient"]["count"] == 3

    visualize_aggregation(aggregated, f"{tmp_path}/fhir-query.html")
    assert pathlib.Path(f"{tmp_path}/fhir-query.html").exists()
    # to see the visualization, cp to tmp
    # shutil.copy(f"{tmp_path}/fhir-query.html", "/tmp/fhir-query.html")

    count = 0
    for _ in db.flattened_specimens():
        count += 1
        print(_)
        assert "patient_id" in _
    assert count == 3, "Expected 3 flattened specimens"



from click.testing import CliRunner
from fhir_query.cli import cli


def test_default_option() -> None:
    """Test default option."""
    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])
    print(result.output)
    output = result.output
    for _ in ["main*", "summarize", "visualize"]:
        assert _ in output, f"Expected {_} in {output}"


def test_help_option() -> None:
    """Test help option."""
    runner = CliRunner()
    result = runner.invoke(cli, ["main", "--help"])
    output = result.output
    print(output)
    assert "Usage:" in output
    assert "--fhir-base-url" in output
    assert "--graph-definition-id" in output
    assert "--graph-definition-file-path" in output
    assert "--path" in output
    assert "--db-path" in output
    assert "--dry-run" in output
    assert "--debug" in output
    assert "--log-file" in output


def test_visualize_help() -> None:
    """Test visualizer help."""
    runner = CliRunner()
    result = runner.invoke(cli, ["visualize", "--help"])
    output = result.output
    assert "Usage:" in output
    assert "--db-path" in output
    assert "--output-path" in output



# tests/conftest.py
import logging
import urllib
from typing import Generator, Any

import pytest
import httpx
from httpx import Response
from pytest_httpx import HTTPXMock


@pytest.fixture
def mock_fhir_server(httpx_mock: HTTPXMock) -> Generator[HTTPXMock, Any, Any]:
    def dummy_callback(request: httpx.Request) -> Response:

        logging.warning(f"Request: {request.url.path}, {str(request.url.params)}")

        if request.url.path == "/ResearchStudy" and str(request.url.params) == "_id=123":
            return Response(
                200, json={"resourceType": "Bundle", "entry": [{"resource": {"resourceType": "ResearchStudy", "id": "123"}}]}
            )

        if request.url.path == "/ResearchSubject":
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "entry": [
                        {"resource": {"resourceType": "ResearchSubject", "id": "123RS", "subject": {"reference": "Patient/123"}}},
                    ],
                },
            )

        if request.url.path == "/Patient/123":
            return Response(200, json={"resourceType": "Patient", "id": "123"})

        if request.url.path == "/Patient/999":
            return Response(404, json={"error": "Not found"})

        if (
            request.url.path == "/Patient"
            and str(request.url.params)
            == "_has%3AResearchSubject%3Asubject%3Astudy=ResearchStudy%2F123&_revinclude=ResearchSubject%3Asubject&_revinclude=Group%3Amember&_count=1000&_total=accurate"
        ):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "Patient", "id": "1", "name": [{"family": "Smith", "given": ["John"]}]}},
                        {"resource": {"resourceType": "Patient", "id": "2", "name": [{"family": "Doe", "given": ["Jane"]}]}},
                        {"resource": {"resourceType": "Patient", "id": "3", "name": [{"family": "Brown", "given": ["Charlie"]}]}},
                    ],
                },
            )

        if request.url.path == "/Specimen" and "subject=Patient" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "Specimen", "id": "S1", "subject": {"reference": "Patient/1"}}},
                        {"resource": {"resourceType": "Specimen", "id": "S2", "subject": {"reference": "Patient/2"}}},
                        {"resource": {"resourceType": "Specimen", "id": "S3", "subject": {"reference": "Patient/3"}}},
                    ],
                },
            )

        if request.url.path == "/Group" and "member=Specimen" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "Group", "id": "G1"}},
                    ],
                },
            )

        if request.url.path == "/DocumentReference" and "subject=Group" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "DocumentReference", "id": "DR1"}},
                    ],
                },
            )

        if request.url.path == "/DocumentReference" and "subject=Patient" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "DocumentReference", "id": "DR2"}},
                    ],
                },
            )

        if request.url.path == "/Observation" and "subject=Patient" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "Observation", "id": "O1"}},
                    ],
                },
            )

        if request.url.path == "/Procedure" and "subject=Patient" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "Procedure", "id": "P1"}},
                    ],
                },
            )

        if request.url.path == "/ServiceRequest" and "subject=Patient" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "ServiceRequest", "id": "SR1"}},
                    ],
                },
            )

        if request.url.path == "/ImagingStudy" and "subject=Patient" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "ImagingStudy", "id": "IS1"}},
                    ],
                },
            )

        if request.url.path == "/Condition" and "subject=Patient" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "Condition", "id": "C1"}},
                    ],
                },
            )

        if request.url.path == "/Medication" and "subject=Patient" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "Medication", "id": "M1"}},
                    ],
                },
            )

        if request.url.path == "/MedicationAdministration" and "subject=Patient" in str(request.url.params):
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {
                            "resource": {
                                "resourceType": "MedicationAdministration",
                                "id": "MA1",
                                "medication": {"reference": {"reference": "Medication/M1"}},
                            }
                        },
                    ],
                },
            )

        if request.url.path == "/Medication":
            return Response(
                200,
                json={
                    "resourceType": "Bundle",
                    "type": "searchset",
                    "entry": [
                        {"resource": {"resourceType": "Medication", "id": "M1"}},
                    ],
                },
            )
        # unexpected request
        print(request.url, str(request.url.params))
        assert False, f"Unexpected url:{request.url} path:{request.url.path}, params:{str(request.url.params)}"

    httpx_mock.add_callback(dummy_callback)

    yield httpx_mock



import asyncio
import json
import logging
import sys
from typing import Any

import click
import pandas as pd
import requests
import yaml
from click_default_group import DefaultGroup
from fhir.resources.graphdefinition import GraphDefinition
from halo import Halo

from fhir_query import GraphDefinitionRunner, setup_logging
from fhir_query.dataframer import Dataframer
from fhir_query.visualizer import visualize_aggregation
from fhir_query.vocabulary import vocabulary_simplifier


@click.group(cls=DefaultGroup, default="main")
def cli():
    """Run FHIR GraphDefinition traversal."""
    pass


@cli.command()
@click.option("--fhir-base-url", required=True, help="Base URL of the FHIR server.")
@click.option("--graph-definition-id", help="ID of the GraphDefinition.")
@click.option("--graph-definition-file-path", help="Path to the GraphDefinition JSON file.")
@click.option("--path", required=False, default=None, help="Query to start traversal.")
@click.option("--db-path", default="/tmp/fhir-graph.sqlite", help="path to of sqlite db")
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Perform a dry run without making any changes.",
)
@click.option("--debug", is_flag=True, help="Enable debug mode.")
@click.option("--log-file", default="app.log", help="Path to the log file.")
def main(
    fhir_base_url: str,
    graph_definition_id: str,
    graph_definition_file_path: str,
    path: str,
    debug: bool,
    db_path: str,
    dry_run: bool,
    log_file: str,
) -> None:
    """Run FHIR GraphDefinition traversal."""

    setup_logging(debug, log_file)

    if fhir_base_url.endswith("/"):
        fhir_base_url = fhir_base_url[:-1]

    if not graph_definition_id and not graph_definition_file_path:
        raise click.UsageError("You must provide either --graph-definition-id or --graph-definition-file-path.")

    runner = GraphDefinitionRunner(fhir_base_url, db_path, debug)

    async def run_runner() -> None:
        if graph_definition_file_path:
            with open(graph_definition_file_path, "r") as f:
                if graph_definition_file_path.endswith(".yaml") or graph_definition_file_path.endswith(".yml"):
                    graph_definition = yaml.safe_load(f)
                else:
                    graph_definition = json.load(f)
        else:
            graph_definition = await runner.fetch_graph_definition(graph_definition_id)

        _ = GraphDefinition(**graph_definition)
        click.echo(f"{_.id} is valid FHIR R5 GraphDefinition", file=sys.stderr)
        if dry_run:
            click.echo("Dry run mode enabled. Exiting.", file=sys.stderr)
            exit(0)

        logging.debug(runner.db_path)
        spinner = Halo(text=f"Running {_.id} traversal", spinner="dots", stream=sys.stderr)
        try:
            await runner.run(graph_definition, path, spinner)
        finally:
            spinner.stop()
        click.echo(f"Aggregated Results: {runner.count_resource_types()}", file=sys.stderr)
        click.echo(f"database available at: {runner.db_path}", file=sys.stderr)

    try:
        asyncio.run(run_runner())
    except Exception as e:
        logging.error(f"Error: {e}", exc_info=True)
        click.echo(f"Error: {e}", file=sys.stderr)
        if debug:
            raise e


@cli.command()
@click.option("--fhir-base-url", required=True, help="Base URL of the FHIR server.")
@click.option("--raw", is_flag=True, default=False, help="Do not create a dataframe. default=False")
@click.option("--tsv", is_flag=True, default=True, help="Render dataframe as tsv. default=True")
@click.option("--debug", is_flag=True, help="Enable debug mode.")
@click.option("--log-file", default="./fq.log", help='Path to the log file. default="./fq.log"')
@click.argument("output_path", type=click.File("w"), required=False, default=sys.stdout)
def vocabulary(
    fhir_base_url: str,
    output_path: click.File,
    debug: bool,
    log_file: str,
    raw: bool,
    tsv: bool,
) -> None:
    """Retrieve Vocabulary Observation and ResearchStudy resources from the FHIR server.
    \b

    OUTPUT_PATH: Path to the output file. If not provided, the output will be printed to stdout.
    """

    setup_logging(debug, log_file)

    if fhir_base_url.endswith("/"):
        fhir_base_url = fhir_base_url[:-1]

    output_stream: Any = output_path

    try:
        with Halo(text="Collecting vocabularies", spinner="dots", stream=sys.stderr) as spinner:
            query_url = f"{fhir_base_url}/Observation?code=vocabulary&_include=Observation:focus"
            response = requests.get(query_url, timeout=300)
            response.raise_for_status()
            bundle = response.json()
            results = bundle
            if not raw:
                results = vocabulary_simplifier(bundle)
            if not tsv:
                yaml_results = yaml.dump(results, default_flow_style=False, sort_keys=False)
                print(yaml_results, file=output_stream)
            else:
                df = pd.DataFrame(results)
                df.to_csv(output_stream, sep="\t", index=False)
            spinner.succeed(f"Wrote {len(results)} vocabularies to {output_stream.name}")

    except Exception as e:
        logging.error(f"Error: {e}", exc_info=True)
        click.echo(f"Error: {e}", file=sys.stderr)
        if debug:
            raise e


@cli.command(name="visualize")
@click.option("--db-path", default="/tmp/fhir-graph.sqlite", help="path to sqlite db")
@click.option("--output-path", default="/tmp/fhir-graph.html", help="path output html")
@click.option(
    "--ignored-edges", "-i", multiple=True, help="Edges to ignore in the visualization", default=["member", "part-of-study"]
)
def visualize(db_path: str, output_path: str, ignored_edges: list[str]) -> None:
    """Visualize the aggregation results."""
    from fhir_query import ResourceDB

    try:
        db = ResourceDB(db_path=db_path)
        visualize_aggregation(db.aggregate(ignored_edges), output_path)
        click.echo(f"Wrote: {output_path}", file=sys.stderr)
    except Exception as e:
        logging.error(f"Error: {e}", exc_info=True)
        click.echo(f"Error: {e}", file=sys.stderr)
        # raise e


@cli.command(name="summarize")
@click.option("--db-path", default="/tmp/fhir-graph.sqlite", help="path to sqlite db")
def summarize(db_path: str) -> None:
    """Summarize the aggregation results."""
    from fhir_query import ResourceDB

    try:
        db = ResourceDB(db_path=db_path)
        yaml.dump(json.loads(json.dumps(db.aggregate())), sys.stdout, default_flow_style=False)

    except Exception as e:
        logging.error(f"Error: {e}", exc_info=True)
        click.echo(f"Error: {e}", file=sys.stderr)
        # raise e


@cli.command(name="dataframe")
# TODO - fix the default paths
@click.option("--db-path", default="/tmp/fhir-graph.sqlite", help="path to sqlite db")
@click.option("--output-path", default="/tmp/fhir-graph.tsv", help="path output tsv")
@click.option(
    "--dtale",
    "launch_dtale",
    default=False,
    show_default=True,
    is_flag=True,
    help="Open the graph in a browser using the dtale package for interactive data exploration.",
)
@click.argument(
    "data_type",
    required=True,
    type=click.Choice(["Specimen", "DocumentReference", "ResearchSubject", "Patient"]),
    default="Specimen",
)
def dataframe(db_path: str, output_path: str, launch_dtale: bool, data_type: str) -> None:
    """Create dataframes from the local db."""

    try:
        db = Dataframer(db_path=db_path)
        # TODO - add more data types - including condition
        assert data_type in ["Specimen", "Patient"], f"Sorry {data_type} dataframe is not supported yet."

        df: pd.DataFrame | None = None
        if data_type == "Specimen":
            df = pd.DataFrame(db.flattened_specimens())
        if data_type == "Patient":
            df = pd.DataFrame(db.flattened_patients())

        if launch_dtale:
            # TODO - add check that dtale is installed
            import dtale

            dtale.show(df, subprocess=False, open_browser=True, port=40000)
        elif df is not None:
            # export to csv
            file_name = output_path if output_path else f"{data_type}.csv"
            df.to_csv(file_name, index=False)
            click.secho(f"Saved {file_name}", file=sys.stderr)
        else:
            click.secho(f"No data found for {data_type}", file=sys.stderr)

    except Exception as e:
        logging.error(f"Error: {e}", exc_info=True)
        click.echo(f"Error: {e}", file=sys.stderr)
        # raise e


if __name__ == "__main__":
    cli()



from fhir_query import find_key_with_path, get_value_from_path


def test_recursive_search():
    data = {"a": {"b": {"c": 42}, "d": [{"e": 100}, {"f": 200}], "g": {"h": {"f": 300}}}}

    key_to_find = "f"
    result = find_key_with_path(data, key_to_find)
    expected = [{"path": ["a", "d", 1, "f"], "value": 200}, {"path": ["a", "g", "h", "f"], "value": 300}]
    assert result == expected, result

    ignored_keys = ["g"]
    result = find_key_with_path(data, key_to_find, ignored_keys=ignored_keys)
    expected = [{"path": ["a", "d", 1, "f"], "value": 200}]
    assert result == expected, result


def test_recursive_search_empty():
    data = {"a": {"b": {"c": 42}, "d": [{"e": 100}, {"f": 200}], "g": {"h": {"f": 300}}}}

    key_to_find = "x"
    result = find_key_with_path(data, key_to_find)
    expected = []
    assert result == expected, result


def test_get_value_from_path():
    data = {"a": {"b": {"c": 42}, "d": [{"e": 100}, {"f": 200}], "g": {"h": {"f": 300}}}}
    path = ["a", "d", 1, "f"]
    result = get_value_from_path(data, path)
    expected = 200
    assert result == expected, result



from pyvis.network import Network

from fhir_query import ResourceDB


def _container():
    """Create a pyvis container."""
    return Network(notebook=True, cdn_resources="in_line")  # filter_menu=True, select_menu=True


def _load(net: Network, aggregation: dict) -> Network:
    """Load the aggregation into the visualization network."""
    # add vertices
    for resource_type, _ in aggregation.items():
        assert "count" in _, _
        net.add_node(resource_type, label=f"{resource_type}/{_['count']}")
    # add edges
    for resource_type, _ in aggregation.items():
        for ref in _.get("references", {}):
            count = _["references"][ref]["count"]
            if resource_type not in net.get_nodes():
                net.add_node(resource_type, label=f"{resource_type}/?")
            if ref not in net.get_nodes():
                net.add_node(ref, label=f"{ref}/?")
            net.add_edge(resource_type, ref, title=count, value=count)
    return net


def visualize_aggregation(aggregation: dict, output_path: str) -> None:
    """Visualize the aggregation."""
    # Load it into a pyvis
    net = _load(_container(), aggregation)
    net.show_buttons(filter_=["physics"])
    net.save_graph(str(output_path))


def create_network_graph(db_path: str, output_path: str) -> None:
    """Render metadata as a network graph into output_path.

    \b
    db_path: The directory path to the db.
    output_path: The path to save the network graph.
    """
    db = ResourceDB(db_path=db_path)
    visualize_aggregation(db.aggregate(), output_path)



import json
from collections import defaultdict
from functools import lru_cache
from typing import Generator

from fhir_query import ResourceDB

import inflection

from pydantic import BaseModel, computed_field
from typing import Dict, List, Optional, Tuple


#######################
# FHIR HELPER METHODS #
#######################


def get_nested_value(d: dict, keys: list):
    for key in keys:
        try:
            d = d[key]
        except (KeyError, IndexError, TypeError):
            return None
    return d


def normalize_coding(resource_dict: Dict) -> List[Tuple[str, str]]:
    """normalize any nested coding"""

    def extract_coding(coding_list):
        # return a concatenated string
        # or alternatively return an array
        return [coding.get("display", coding.get("code", "")) for coding in coding_list]

    def find_codings_in_dict(d: dict, parent_key: str = "") -> list[tuple[str, str]]:  # TODO - parent_key not used?
        codings = []
        for key, value in d.items():
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        # Check if the dict contains a 'coding' list
                        if "coding" in item and isinstance(item["coding"], list):
                            coding_string = extract_coding(item["coding"])
                            codings.append((coding_string, key))
                        if "code" in item:
                            coding_string = item.get("display", item.get("code"))
                            codings.append((coding_string, key))

                        # Recursively search in the dict
                        codings.extend(find_codings_in_dict(item, key))
            elif isinstance(value, dict):
                # Check if the dict contains a 'coding' list
                if "coding" in value and isinstance(value["coding"], list):
                    coding_string = extract_coding(value["coding"])
                    codings.append((coding_string, key))

                # Recursively search in the dict
                codings.extend(find_codings_in_dict(value, key))
        return codings

    return find_codings_in_dict(resource_dict)


def normalize_value(resource_dict: dict) -> tuple[Optional[str], Optional[str]]:
    """return a tuple containing the normalized value and the name of the field it was derived from"""

    if "valueQuantity" in resource_dict:
        value = resource_dict["valueQuantity"]
        value_normalized = f"{value['value']} {value.get('unit', '')}"
        value_source = "valueQuantity"
    elif "valueCodeableConcept" in resource_dict:
        value = resource_dict["valueCodeableConcept"]
        value_normalized = " ".join([coding.get("display", coding.get("code", "")) for coding in value.get("coding", [])])
        value_source = "valueCodeableConcept"
    elif "valueCoding" in resource_dict:
        value = resource_dict["valueCoding"]
        value_normalized = value["display"]
        value_source = "valueCoding"
    elif "valueString" in resource_dict:
        value_normalized = resource_dict["valueString"]
        value_source = "valueString"
    elif "valueCode" in resource_dict:
        value_normalized = resource_dict["valueCode"]
        value_source = "valueCode"
    elif "valueBoolean" in resource_dict:
        value_normalized = str(resource_dict["valueBoolean"])
        value_source = "valueBoolean"
    elif "valueInteger" in resource_dict:
        value_normalized = str(resource_dict["valueInteger"])
        value_source = "valueInteger"
    elif "valueRange" in resource_dict:
        value = resource_dict["valueRange"]
        low = value["low"]
        high = value["high"]
        value_normalized = f"{low['value']} - {high['value']} {low.get('unit', '')}"
        value_source = "valueRange"
    elif "valueRatio" in resource_dict:
        value = resource_dict["valueRatio"]
        numerator = value["numerator"]
        denominator = value["denominator"]
        value_normalized = f"{numerator['value']} {numerator.get('unit', '')}/{denominator['value']} {denominator.get('unit', '')}"
        value_source = "valueRatio"
    elif "valueSampledData" in resource_dict:
        value = resource_dict["valueSampledData"]
        value_normalized = value["data"]
        value_source = "valueSampledData"
    elif "valueTime" in resource_dict:
        value_normalized = resource_dict["valueTime"]
        value_source = "valueTime"
    elif "valueDateTime" in resource_dict:
        value_normalized = resource_dict["valueDateTime"]
        value_source = "valueDateTime"
    elif "valuePeriod" in resource_dict:
        value = resource_dict["valuePeriod"]
        value_normalized = f"{value['start']} to {value['end']}"
        value_source = "valuePeriod"
    elif "valueUrl" in resource_dict:
        value_normalized = resource_dict["valueUrl"]
        value_source = "valueUrl"
    elif "valueDate" in resource_dict:
        value_normalized = resource_dict["valueDate"]
        value_source = "valueDate"
    elif "valueCount" in resource_dict:
        value_normalized = resource_dict["valueCount"]["value"]
        value_source = "valueCount"
    elif "valueReference" in resource_dict:
        value_normalized = resource_dict["valueReference"]["reference"]
        value_source = "valueReference"

    else:
        value_normalized, value_source = None, None
        # for debugging...
        # raise ValueError(f"value[x] not found in {resource_dict}")

    return value_normalized, value_source


def normalize_for_guppy(key: str):
    """normalize a key so that it can be loaded into Guppy as a column name"""
    guppy_table = str.maketrans(
        {
            ".": "",
            " ": "_",
            "[": "",
            "]": "",
            "'": "",
            ")": "",
            "(": "",
            ",": "",
            "/": "_per_",
            "-": "to",
            "#": "number",
            "+": "_plus_",
            "%": "percent",
            "&": "_and_",
        }
    )
    return key.translate(guppy_table)


def traverse(resource):
    """simplify a resource's fields, returned as a dict of values,
    where keys are prefixed with "resourceType_" """

    final_subject = {}
    simplified_subject = SimplifiedResource.build(resource=resource).simplified
    prefix = simplified_subject["resourceType"].lower()
    for k, v in simplified_subject.items():
        if k in ["resourceType"]:
            continue
        final_subject[f"{prefix}_{k}"] = v

    return final_subject


########################
# FHIR BUILDER OBJECTS #
########################


class SimplifiedFHIR(BaseModel):
    """All Simplifiers should inherit from this class."""

    warnings: list[str] = []
    """A list of warnings generated during the simplification process."""
    resource: dict
    """The FHIR resource to be simplified."""

    @computed_field()  # type: ignore[prop-decorator]
    @property
    def simplified(self) -> dict:
        _ = self.identifiers.copy() if self.identifiers else {}
        _.update(self.scalars)
        _.update(self.codings)
        _.update(self.extensions)
        _.update(self.values)
        return _

    def simplify_extensions(self, resource: dict = None, _extensions: dict = None) -> dict:  # type: ignore  # noqa
        """Extract extension values, derive key from extension url"""

        def _populate_simplified_extension(extension: dict):
            # simple extension
            value_normalized, extension_key = normalize_value(extension)
            extension_key = extension["url"].split("/")[-1]
            extension_key = inflection.underscore(extension_key).removesuffix(".json").removeprefix("structure_definition_")  # type: ignore[arg-type]
            if value_normalized is None:
                pass
            assert value_normalized is not None, f"extension: {extension_key} = {value_normalized} {extension}"
            _extensions[extension_key] = value_normalized

        if not _extensions:
            _extensions = {}

        if not resource:
            resource = self.resource

        if "extension" not in resource.keys():
            return _extensions

        for _ in resource.get("extension", [resource]):
            if "extension" not in _.keys():
                if "resourceType" not in _.keys():
                    _populate_simplified_extension(_)
                continue
            elif set(_.keys()) == {"url", "extension"}:
                for child_extension in _["extension"]:
                    self.simplify_extensions(resource=child_extension, _extensions=_extensions)

        return _extensions

    @computed_field  # type: ignore[prop-decorator]
    @property
    def extensions(self) -> dict:
        return self.simplify_extensions()

    @computed_field  # type: ignore[prop-decorator]
    @property
    def scalars(self) -> dict:
        """Return a dictionary of scalar values."""
        return {k: v for k, v in self.resource.items() if (not isinstance(v, list) and not isinstance(v, dict))}

    @computed_field  # type: ignore[prop-decorator]
    @property
    def codings(self) -> dict:
        """Return a dictionary of scalar values."""
        _codings = {}
        for k, v in self.resource.items():
            # these are handled in separate methods
            if k in ["identifier", "extension", "component", "code"]:
                continue
            elif isinstance(v, list):
                for elem in v:
                    if isinstance(elem, dict):
                        # TODO: implement hierarchy of codes rather than just taking last code?
                        for value, source in normalize_coding(elem):
                            if len(v) > 1 and get_nested_value(elem, [source, 0, "system"]):
                                _codings[elem[source][0]["system"].split("/")[-1]] = value
                            else:
                                _codings[k] = value
            elif isinstance(v, dict):
                for value, elem in normalize_coding(v):
                    _codings[k] = value

        return _codings

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifiers(self) -> dict:
        """Return the first of a resource and any other resources"""
        identifiers = self.resource.get("identifier", [])
        identifiers_len = len(identifiers)

        if not identifiers_len:
            return {"identifier": None}
        elif identifiers_len == 1:
            return {"identifier": identifiers[0].get("value")}
        else:
            base_identifier = {"identifier": identifiers[0].get("value")}
            base_identifier.update(
                {identifier.get("system").split("/")[-1]: identifier.get("value") for identifier in identifiers[1:]}
            )
            return base_identifier

    @computed_field  # type: ignore[prop-decorator]
    @property
    def values(self) -> dict:
        """Return a dictionary of source:value."""
        # FIXME: values that are scalars are processed twice: once in scalars once here in values (e.g. valueString)
        value, source = normalize_value(self.resource)
        if not value:
            return {}

        # update the key if code information is available

        if self.resource.get("code", {}).get("text", None):
            source = self.resource["code"]["text"]
        else:
            source = self.resource["code"]["coding"][0].get("display", self.resource["code"]["coding"][0].get("code"))
        if not source:
            source = "NA"
        source = inflection.underscore(inflection.parameterize(source))
        return {source: value}


class SimplifiedObservation(SimplifiedFHIR):
    @computed_field  # type: ignore[prop-decorator]
    @property
    def codings(self) -> dict:
        """does everything but gets rid of code since that's dealt with in values"""
        _codings = super().codings
        if "code" in _codings:
            del _codings["code"]

        return _codings

    # TODO: remove after data fix
    @computed_field  # type: ignore[prop-decorator]
    @property
    def scalars(self) -> dict:
        """Return a dictionary of scalar values."""
        return {
            k: v
            for k, v in self.resource.items()
            if (not isinstance(v, list) and not isinstance(v, dict) and not k == "valueString")
        }

    @computed_field  # type: ignore[prop-decorator]
    @property
    def values(self) -> dict:
        """Return a dictionary of 'value':value or <component>:value.
        https://build.fhir.org/observation-definitions.html#Observation.component
        """

        # get top-level value in dict if it exists
        _values = super().values
        if len(_values) == 0:
            assert "component" in self.resource, "no component nor top-level value found"

            # get component codes
            if "component" in self.resource:
                for component in self.resource["component"]:
                    value, source = normalize_value(component)
                    if component.get("code", {}).get("text", None):
                        source = component["code"]["text"]
                    if not value:
                        continue
                    _values[source] = value

            # knowing there's now at least 1 item in _values
            if "component" in self.resource:
                # ensure no top-level value is not duplicating a component code value
                # TODO: ensure this value_key corresponds to percent_tumor on some runs due to getting display
                value_key = [k for k in _values][0]
                assert (
                    value_key not in self.resource["component"]
                ), """duplicate code value found, only specify the code value in the component, see Rule obs-7
                    https://build.fhir.org/observation.html#invs"""

                # get component codes
                if "component" in self.resource:
                    for component in self.resource["component"]:
                        value, source = normalize_value(component)
                        if component.get("code", {}).get("text", None):
                            source = component["code"]["text"]
                        if not value:
                            continue
                        _values[source] = value

        if "code" in self.resource and "text" in self.resource["code"]:
            _values["observation_code"] = self.resource["code"]["text"]

        assert not [_ for _ in _values.keys() if _.startswith("value")], f"key misnamed {_values}\n  {self.resource}"
        assert len(_values) > 0, f"no values found in Observation: {self.resource}"

        return _values


class SimplifiedDocumentReference(SimplifiedFHIR):
    @computed_field  # type: ignore[prop-decorator]
    @property
    def values(self) -> dict:
        """Return a dictionary of 'value':value."""
        _values = super().values
        for content in self.resource.get("content", []):
            if "attachment" in content:
                for k, v in SimplifiedFHIR(resource=content["attachment"]).simplified.items():
                    if k in ["identifier", "extension"]:
                        continue
                    _values[k] = v
        return _values


class SimplifiedCondition(SimplifiedFHIR):
    @computed_field  # type: ignore[prop-decorator]
    @property
    def codings(self) -> dict:
        # only go through the work if code exists
        if "code" not in self.resource:
            return {}

        # get field name
        codings_dict = super().codings
        if "cate.g.ory" in codings_dict:
            key = codings_dict["cate.g.ory"]
            del codings_dict["cate.g.ory"]
        else:
            key = "code"

        # TODO: implement hierarchy of codes rather than just taking last code?
        value, _ = normalize_coding(self.resource["code"])[-1]
        return {key: value}


class SimplifiedResource(object):
    """A simplified FHIR resource, a factory method."""

    @staticmethod
    def build(resource: dict) -> SimplifiedFHIR:
        """Return a simplified FHIR resource."""

        resource_type = resource.get("resourceType", None)
        if resource_type == "Observation":
            return SimplifiedObservation(resource=resource)
        if resource_type == "DocumentReference":
            return SimplifiedDocumentReference(resource=resource)
        if resource_type == "Condition":
            return SimplifiedCondition(resource=resource)
        return SimplifiedFHIR(resource=resource)


class Dataframer(ResourceDB):
    def __init__(self, db_path: str):
        super().__init__(db_path)

    def get_subject(self, resource: dict) -> dict:
        """get the resource's subject field if it exists"""

        # ensure resource has subject field
        subject_key = get_nested_value(resource, ["subject", "reference"])
        if subject_key is None:
            return {}

        # traverse the resource of the subject and return its values
        cursor = self.connection.cursor()
        cursor.execute("SELECT * FROM resources WHERE key = ?", (subject_key,))
        row = cursor.fetchone()
        assert row, f"{subject_key} not found in database"
        _, _, _, raw_subject = row
        subject = json.loads(raw_subject)
        return traverse(subject)

    def get_resources_by_reference(self, resource_type: str, reference_field: str, reference_type: str) -> dict[str, list]:
        """Given a set of resources of type resource_type, map each unique reference in reference field of type reference_type to its associated resources
        ex: use all Observations with a Specimen focus, map Specimen IDs to its list of associated Observations and return the map
        """

        # ensure reference field is allowed
        allowed_fields = ["focus", "subject", "specimen", "basedOn"]
        assert reference_field in allowed_fields, f"Field not implemented, choose between {allowed_fields}"

        cursor = self.connection.cursor()
        cursor.execute(
            """
            SELECT *
            FROM resources
            WHERE resource_type = ?
        """,
            (resource_type,),
        )

        resource_by_reference_id = defaultdict(list)

        for _, _, _, raw_resource in cursor.fetchall():
            resource = json.loads(raw_resource)

            # determine which how to process the field
            if reference_field == "focus" and "focus" in resource:
                # add the resource (e.g. observation) for each focus reference to the dict
                for i in range(len(resource["focus"])):
                    reference_key: str = get_nested_value(resource, [reference_field, i, "reference"])
                    if reference_key is not None and reference_type in reference_key:
                        reference_id = reference_key.split("/")[-1]
                        resource_by_reference_id[reference_id].append(resource)

            if reference_field == "specimen" and "specimen" in resource:
                # add the resource (e.g. observation) for each specimen reference to the dict
                for i in range(len(resource["specimen"])):
                    reference_key = get_nested_value(resource, [reference_field, i, "reference"])
                    if reference_key is not None and reference_type in reference_key:
                        reference_id = reference_key.split("/")[-1]
                        resource_by_reference_id[reference_id].append(resource)

            if reference_field == "basedOn" and "basedOn" in resource:
                # add the resource (e.g. observation) for each basedOn reference to the dict
                for i in range(len(resource["basedOn"])):
                    reference_key = get_nested_value(resource, [reference_field, i, "reference"])
                    if reference_key is not None and reference_type in reference_key:
                        reference_id = reference_key.split("/")[-1]
                        resource_by_reference_id[reference_id].append(resource)

            elif reference_field == "subject":
                # add the resource (e.g. observation) to the dict
                reference_key = get_nested_value(resource, [reference_field, "reference"])
                if reference_key is not None and reference_type in reference_key:
                    reference_id = reference_key.split("/")[-1]
                    resource_by_reference_id[reference_id].append(resource)

        return resource_by_reference_id

    def get_observations_by_focus(self, focus_type: str) -> dict[str, list]:
        """get all Observations that have a focus of resource type focus_type"""
        return self.get_resources_by_reference("Observation", "focus", focus_type)

    @lru_cache(maxsize=None)
    def flattened_specimens(self) -> Generator[dict, None, None]:
        """generator that yields specimens populated with `Specimen.subject` fields
        and Observation codes through `Observation.focus`"""

        resource_type = "Specimen"
        cursor = self.connection.cursor()

        # get a dict mapping focus ID to its associated observations
        specimen_observations_by_focus_id = self.get_observations_by_focus(resource_type)
        service_requests_by_specimen_id = self.get_resources_by_reference("ServiceRequest", "specimen", "Specimen")
        document_references_by_based_on_id = self.get_resources_by_reference("DocumentReference", "basedOn", "ServiceRequest")

        # flatten each document reference
        cursor.execute("SELECT * FROM resources where resource_type = ?", (resource_type,))
        for _, _, _, resource in cursor.fetchall():
            specimen = json.loads(resource)
            yield self.flattened_specimen(
                specimen, specimen_observations_by_focus_id, service_requests_by_specimen_id, document_references_by_based_on_id
            )

    def flattened_specimen(
        self, specimen: dict, observation_by_id: dict, service_requests_by_specimen_id, document_references_by_based_on_id
    ) -> dict:
        """Return the specimen with everything resolved."""

        # create simple specimen dict
        flat_specimen = traverse(specimen)

        # extract its .subject and append its fields (including id)
        subject = self.get_subject(specimen)
        if "patient_id" in subject:
            assert len(self.flattened_patients()) > 1, f"Length of flattened_patients is {len(self.flattened_patients())}"
            _flattened_patient = next(
                iter([_ for _ in self.flattened_patients() if _["patient_id"] == subject["patient_id"]]), None
            )
            if not _flattened_patient:
                print(f"Patient not found {subject['patient_id']} {[_['patient_id'] for _ in self.flattened_patients()]}")
            else:
                subject = {f"patient_{k}".replace("patient_patient_", "patient_"): v for k, v in _flattened_patient.items()}
        flat_specimen.update(subject)

        # populate observation codes for each associated observation
        if specimen["id"] in observation_by_id:
            observations = observation_by_id[specimen["id"]]
            # TODO: assumes there are no duplicate column names in each observation
            for observation in observations:
                flat_observation = SimplifiedResource.build(resource=observation).values
                flat_observation = {f"observation_{k}": v for k, v in flat_observation.items()}
                flat_specimen.update(flat_observation)

        if specimen["id"] in service_requests_by_specimen_id:
            service_requests = service_requests_by_specimen_id[specimen["id"]]
            # TODO: assumes there are no duplicate column names in each observation
            for service_request in service_requests:
                flat_specimen.update(traverse(service_request))
                if service_request["id"] in document_references_by_based_on_id:
                    document_references = document_references_by_based_on_id[service_request["id"]]
                    for document_reference in document_references:
                        flat_specimen.update(traverse(document_reference))

        return flat_specimen

    @lru_cache(maxsize=None)
    def flattened_patients(self) -> list[dict]:
        """
        Generator that yields flattened Patient records.
        Each flattened Patient merges in fields from:
            - Observations that reference the Patient via the focus field
        """
        resource_type = "Patient"
        cursor = self.connection.cursor()

        observations_by_focus = self.get_resources_by_reference("Observation", "focus", "Patient")
        cursor.execute("SELECT * FROM resources WHERE resource_type = ?", (resource_type,))
        _flattened_patients = []
        for _, _, _, resource in cursor.fetchall():
            patient = json.loads(resource)
            _flattened_patients.append(self.flattened_patient(patient, observations_by_focus))
        return _flattened_patients

    @staticmethod
    def flattened_patient(patient: dict, observations_by_subject: dict) -> dict:
        """Return the flattened Patient record with related Observations"""
        flat_patient = traverse(patient)

        if patient["id"] in observations_by_subject:
            observations = observations_by_subject[patient["id"]]
            for observation in observations:
                flat_observation = SimplifiedResource.build(resource=observation).values
                flat_observation = {f"observation_{k}": v for k, v in flat_observation.items()}
                flat_patient.update(flat_observation)

        return flat_patient



from setuptools import setup, find_packages


def parse_requirements(filename: str) -> list[str]:
    with open(filename, "r") as file:
        return [line.strip() for line in file if line.strip() and not line.startswith("#")]


setup(
    name="fhir_aggregator_client",
    version="0.1.2",
    packages=find_packages(),
    install_requires=parse_requirements("requirements.txt"),
    extras_require={
        'dtale': ['dtale'],
    },
    entry_points={
        "console_scripts": [
            "fq=fhir_query:cli.cli",
        ],
    },
)
